-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;   
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- possible generics values: hs = {GROESTL_DATA_SIZE_SMALL, GROESTL_DATA_SIZE_BIG}
-- rom_style = {DISTRIBUTED, COMBINATIONAL}
-- all combinations are allowed 

entity groestl_pq_folded is
generic (
	N:integer := GROESTL_DATA_SIZE_SMALL; 
	ROM_STYLE : integer := DISTRIBUTED;
	FF: integer := 2
);
port( 
	clk				: in std_logic;
	rst				: in std_logic;
	p_mode			: in std_logic;	
	sel				: in std_logic_vector(log2(FF)-1 downto 0);
	wr_shiftreg		: in std_logic;
	round			: in std_logic_vector(7 downto 0);
	input 			: in std_logic_vector(n-1 downto 0);
   	output 			: out std_logic_vector(n/FF-1 downto 0));
end groestl_pq_folded;  	  	 	   

architecture round3_pipelined of groestl_pq_folded is
	signal addcons					: std_logic_vector(n-1 downto 0);	 
	signal after_srp				: std_logic_vector(n-1 downto 0);
	signal after_srq				: std_logic_vector(n-1 downto 0);
	signal after_shiftrow			: std_logic_vector(n-1 downto 0);
	signal after_shiftrow_reg 		: std_logic_vector(n-n/ff-1 downto 0);
	signal after_shiftrow_select 	: std_logic_vector(n/ff-1 downto 0);
	signal after_subbyte			: std_logic_vector(n/ff-1 downto 0);
	signal after_reg				: std_logic_vector(n/ff-1 downto 0);																 
	constant zero					: std_logic_vector(n-1 downto 0):=(others=>'0'); 
begin
			   
	addconsGen: 
		for i in 0 to n/64-1 generate
			addcons(i*64+63 downto i*64+ 56) <= input(i*64+63 downto i*64+56) xor round xor (conv_std_logic_vector(n/64-1-i,4) & x"0") when (p_mode = '1') else (not input(i*64+63 downto i*64+ 56));
			addcons(i*64+55 downto i*64+  8) <= input(i*64+55 downto i*64+ 8) when (p_mode = '1') else (not input(i*64+55 downto i*64+ 8));
			addcons(i*64+ 7 downto i*64+  0) <= input(i*64+ 7 downto i*64+ 0) when (p_mode = '1') else ((conv_std_logic_vector((16-n/64)+i,4) & x"F" ) xor input(i*64+ 7 downto i*64+ 0) xor round);			
		end generate;			   
	
	srp			:entity work.groestl_shiftrow(groestl_shiftrowp) generic map (n=>n)port map (input=>addcons, output=>after_srp);
	srq			:entity work.groestl_shiftrow(groestl_shiftrowq) generic map (n=>n)port map (input=>addcons, output=>after_srq);
	
	after_shiftrow <= after_srp when p_mode='1' else after_srq;	
	
	shiftRowRegInst: 
		regn
		generic map(N=>n-n/ff, init=>zero(n-n/ff-1 downto 0))
		port map (clk => clk, rst => rst, en => wr_shiftreg, input => after_shiftrow(n-n/ff-1 downto 0), output => after_shiftrow_reg );	
	
	
	Sel1_2Gen: 
        if ff = 2 generate
            after_shiftrow_select <= after_shiftrow(n-1 downto n/2) when sel(0) = '0' else after_shiftrow_reg(n/2-1 downto 0);
        end generate;
    Sel1_4Gen: 
        if ff = 4 generate
            with sel(1 downto 0) select
            after_shiftrow_select <= after_shiftrow(n-1 	downto n*3/4) when "00",
		                             after_shiftrow_reg(n*3/4-1 downto n*2/4) when "01",
		                             after_shiftrow_reg(n*2/4-1 downto   n/4) when "10",
		                             after_shiftrow_reg(n/4-1 	downto 	   0) when others;		  
        end generate;
    Sel1_8Gen: 
        if ff = 8 generate
            with sel(2 downto 0) select
            after_shiftrow_select <= after_shiftrow(n-1 	downto n*7/8) when "000",
		                             after_shiftrow_reg(n*7/8-1 downto n*6/8) when "001",
		                             after_shiftrow_reg(n*6/8-1 downto n*5/8) when "010",
		                             after_shiftrow_reg(n*5/8-1 downto n*4/8) when "011",
		                             after_shiftrow_reg(n*4/8-1 downto n*3/8) when "100",
 		                             after_shiftrow_reg(n*3/8-1 downto n*2/8) when "101",
		                             after_shiftrow_reg(n*2/8-1 downto   n/8) when "110",
		                             after_shiftrow_reg(n/8-1 	downto 	   0) when others;	 
        end generate;
	
	sbox_gen: for i in 0 to (n/ff)/AES_SBOX_SIZE - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>after_shiftrow_select(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	
	
	--interimRegInst : sipo
--		generic map ( N => n, M =>n/ff)
--		port map (clk => clk, en => VCC, input => after_subbyte_select, output => after_reg);				  
	
	interimRegInst : 
		entity work.shfreg(struct)
		generic map ( width => N/FF, depth => FF)
		port map (clk => clk, en => VCC, i => after_subbyte, o => after_reg);				  
		
		  
	mc:
		entity work.groestl_mixbytes(groestl_mixbytes)	
		generic map (n=>n/ff)
		port map (input=>after_reg,  output=>output);	

end round3_pipelined; 			 	   
